;; This file contains things that must be fast.  It can be compiled with
;; interrupt checks disabled (but there is no way to specify that).

(declare (usual-integrations))

(declare (integrate-external "sim"))	; for store slots

(declare
 (ignore-reference-traps
  (set resume/r resume/0 resume/1 resume/2 resume/3 resume/4 resume/n)))

(define (call/r val fuel store k)
  (if (fix:< fuel 0)
      (let ((thread (vector-ref store store:thread)))
	(set-thread.context-method! thread resume/r)
	(set-thread.context/1! thread k)
	(set-thread.context/2! thread val)
	(vector-set! store store:thread (thread.next thread))
	fuel)
      (k val fuel store)))

(define (save/r val fuel store k)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/r)
    (set-thread.context/1! thread k)
    (set-thread.context/2! thread val)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/r store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)	; k
     (thread.context/2 thread)	; val
     fuel
     store)))


(define (call/0 k fuel store          p)
  (if (fix:< fuel 0)
      (let ((thread (vector-ref store store:thread)))
	(set-thread.context-method! thread resume/0)
	(set-thread.context/1! thread p)
	(set-thread.context/2! thread k)
	(vector-set! store store:thread (thread.next thread))
	fuel)
      (p k fuel store)))

(define (save/0 k fuel store          p)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/0)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/0 store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store)))


(define (call/1 k fuel store a         p)
  (declare (integrate-operator save/1))
  (if (fix:< fuel 0)
      (save/1 k fuel store a p)
      (p k fuel store a)))

(define (save/1 k fuel store a         p)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/1)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (set-thread.context/3! thread a)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/1 store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store
     (thread.context/3 thread))))	; a


(define (call/2 k fuel store a b         p)
  (declare (integrate-operator save/2))
  (if (fix:< fuel 0)
      (save/2 k fuel store a b p)
      (p k fuel store a b)))

(define (save/2 k fuel store a b         p)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/2)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (set-thread.context/3! thread a)
    (set-thread.context/4! thread b)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/2 store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store
     (thread.context/3 thread)		; a
     (thread.context/4 thread))))	; b


(define (call/3 k fuel store a b c       p)
  (declare (integrate-operator save/3))
  (if (fix:< fuel 0)
      (save/3 k fuel store a b c p)
      (p k fuel store a b c)))

(define (save/3 k fuel store a b c       p)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/3)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (set-thread.context/3! thread a)
    (set-thread.context/4! thread b)
    (set-thread.context/5! thread c)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/3 store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store
     (thread.context/3 thread)		; a
     (thread.context/4 thread)		; b
     (thread.context/5 thread))))	; c


(define (call/4 k fuel store a b c d     p)
  (declare (integrate-operator save/4))
  (if (fix:< fuel 0)
      (save/4 k fuel store a b c d     p)
      (p k fuel store a b c d)))

(define (save/4 k fuel store a b c d     p)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/4)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (set-thread.context/3! thread a)
    (set-thread.context/4! thread b)
    (set-thread.context/5! thread c)
    (set-thread.context/6! thread d)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/4 store fuel)
  (let ((thread (vector-ref store store:thread)))
    ((thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store
     (thread.context/3 thread)		; a
     (thread.context/4 thread)		; b
     (thread.context/5 thread)		; c
     (thread.context/6 thread))))	; d


(define (call/n p k fuel store . args)
  (if (fix:< fuel 0)
      (let ((thread (vector-ref store store:thread)))
	(set-thread.context-method! thread resume/n)
	(set-thread.context/1! thread p)
	(set-thread.context/2! thread k)
	(set-thread.context/3! thread args)
	(vector-set! store store:thread (thread.next thread))
	fuel)
      (apply p k fuel store args)))

(define (save/n p k fuel store . args)
  (let ((thread (vector-ref store store:thread)))
    (set-thread.context-method! thread resume/n)
    (set-thread.context/1! thread p)
    (set-thread.context/2! thread k)
    (set-thread.context/3! thread args)
    (vector-set! store store:thread (thread.next thread))
    fuel))

(define (resume/n store fuel)
  (let ((thread (vector-ref store store:thread)))
    (apply
     (thread.context/1 thread)		; p
     (thread.context/2 thread)		; k
     fuel
     store
     (thread.context/3 thread))))	; args


;; RESUME-PROCESSOR runs the processor until it blocks or runs out of fuel.
;; STEP-PROCESSOR runs the processor like RESUME-PROCESSOR, but keeps
;; the processor clock up to date.  It returns the processor clock
;; time at the end of the step (for use by a scheduler).  So
;; STEP-PROCESSOR should be used by the simulator and RESUME-PROCESSOR
;; by debugging utilities (whcih shoudl keep the processor clock
;; static).
;;
;; Note that STEP-PROCESSOR pre-increments the processor clock, so the
;; correct way to read the processor clock is:
;;
;;    (- (vector-ref store store:counter) fuel)

(define (resume-processor processor fuel)
  (let ((store (processor.store processor)))
    (let ((thread (vector-ref store store:thread)))
      ((thread.context-method thread)
       store fuel))))

(define (step-processor processor fuel)
  (let ((store (processor.store processor)))
    (vector-set! store store:counter (+ (vector-ref store store:counter) fuel))
    (let ((fuel* (resume-processor processor fuel)))
      (let ((clock (- (vector-ref store store:counter) fuel*)))
	(vector-set! store store:counter clock)
	clock))))

(define (global-ref store cache)
  (if (cache.slot cache)
      (vector-ref store (cache.slot cache))
      (global-ref/fault store cache)))

(define (global-set! store cache value)
  (if (cache.slot cache)
      (vector-set! store (cache.slot cache) value)
      (global-set!/fault store cache value)))

(define (global-define! store cache value)
  (if (cache.slot cache)
      (vector-set! store (cache.slot cache) value)
      (if (and (eq? (cache.extra cache) value)
	       (not (eq? value #F)))
	  unspecific
	  (global-define!/fault store cache value))))
