;; Round-robin simulator

(declare (usual-integrations))

(define (simulation.round-robin-scheduler sim #!optional ticks)
  ;; Round-robin until death

  (let ((n (simulation.n-processors sim))
	(v (simulation.processors sim))
	(n-running? (simulation.n-processors sim))
	(ticks (if (default-object? ticks) 10 ticks)))

    (define (resume/halted store fuel)	; fast way to do nothing & use our fuel
      store fuel
      -1)

    (define (continuation:halt value fuel store)
      (processor.termination-notice (store.processor store) fuel value)
      (set! n-running? (- n-running? 1))
      (vector-set! store store:context-method resume/halted)
      fuel)

    (define (step-processor i)
      (let ((processor (vector-ref v i)))
	(let ((store (processor.store processor)))
	  (vector-set! store store:counter
		       (+ (vector-ref store store:counter) ticks))
	  (let ((fuel* (resume-processor processor ticks)))
	    (vector-set! store store:counter
			 (- (vector-ref store store:counter) fuel*))))))

    (define (scheduler-loop)
      (if (> n-running? 0)
	  (do ((i 0 (+ i 1)))
	      ((= i n) (scheduler-loop))
	    (step-processor i))))

    (simulation.reset! sim)
    (simulation.initialize-processor-programs! sim continuation:halt)

    scheduler-loop))

(set! simulation.default-scheduler simulation.round-robin-scheduler)
