;; $Id: simrun2.scm,v 1.4 1997/06/21 22:51:11 adams Exp $

(declare (usual-integrations))

(declare (integrate-external "sim"))

(define (simulation.ag1-scheduler sim #!optional ticks)

  (let* ((n (simulation.n-processors sim))
	 (v (simulation.processors sim))
	 (ticks (if (default-object? ticks) 10 ticks))
	 (processors-remaining n)
	 (agenda (make-agenda))
	 (active-event #F)
	 (end-current-task #F))

    (define (processor-stepper i)
      (define (running-processor)
	(let ((processor (vector-ref v i)))
	  (let ((step-finish-time (step-processor processor ticks)))
	    (agenda.add agenda step-finish-time running-processor))))
      running-processor)

    (define (continuation:halt value fuel store)
      (processor.termination-notice (store.processor store) fuel value)
      (set! processors-remaining (- processors-remaining 1))
      (set! active-event #F)
      (end-current-task #F))

    (define (run)
      (call-with-current-continuation
       (lambda (exit)
	 (set! end-current-task exit)))
      
      (if active-event
	  (if (prompt-for-confirmation
	       (string-append
		"Running event `"
		(write-to-string (lambda-name (procedure-lambda active-event)))
		"'.  Restart (y) or discard (n)"))
	      (active-event)))
      (set! active-event #F)

      (let scheduler-loop ()
	(if (> processors-remaining 0)
	    (begin
	      (let loop ()
		(let ((thunk (agenda.remove-first agenda)))
		  (set! active-event thunk)
		  (thunk)
		  (set! active-event #F)
		  (loop)))
	      (scheduler-loop)))))

    (define metronome
      (let ((time 0))
	(named-lambda (metronome)
	  (cond ((= 0 processors-remaining)
		 (end-current-task #F))
		((agenda.empty? agenda)
		 (error "Agenda is empty but not all processors have finished")
		 (end-current-task #F))
		(else
		 (pp `(time = ,time))
		 (set! time (+ time metronome-period))
		 (agenda.add agenda time metronome))))))

    (simulation.initialize-processor-programs! sim continuation:halt)

    (set-simulation.agenda! sim agenda)
    (do ((i 0 (+ i 1))) ((= i n)) (agenda.add agenda 0 (processor-stepper i)))
    (agenda.add agenda 0 metronome)

    run))

(define metronome-period 1000)

(set! simulation.default-scheduler simulation.ag1-scheduler)
